package de.mdelab.sdm.interpreter.code.debug.ui;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IStorage;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.PlatformObject;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.model.IValue;
import org.eclipse.debug.core.sourcelookup.containers.LocalFileStorage;
import org.eclipse.debug.ui.IDebugModelPresentation;
import org.eclipse.debug.ui.IValueDetailListener;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.jface.viewers.LabelProvider;
import org.eclipse.swt.graphics.Image;
import org.eclipse.ui.IEditorDescriptor;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorRegistry;
import org.eclipse.ui.IPersistableElement;
import org.eclipse.ui.IStorageEditorInput;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.FileEditorInput;

import de.mdelab.sdm.interpreter.code.debug.ui.breakpoints.SDDebugBreakpoint;

public abstract class SDDebugModelPresentation extends LabelProvider implements IDebugModelPresentation {

	private final Map<String, Object> fAttributes;

	public SDDebugModelPresentation() {
		super();
		
		fAttributes = new HashMap<String, Object>( 3 );
	}

	/**
	 * @see IDebugModelPresentation#setAttribute(String, Object)
	 */
	@Override
	public void setAttribute( 	final String p_id,
								final Object p_value ) {
		if ( p_value == null ) {
			fAttributes.remove( p_id );
		}
		else {
			fAttributes.put( p_id, p_value );
		}
	}

	/**
	 * @see IDebugModelPresentation#getEditorId(IEditorInput, Object)
	 */
	@Override
	public String getEditorId( 	final IEditorInput p_input, 
								final Object p_inputObject ) {
		if ( p_inputObject instanceof SDDebugBreakpoint ) {
			return getEditorId( (SDDebugBreakpoint) p_inputObject );
		}
		
		final IEditorRegistry registry = PlatformUI.getWorkbench().getEditorRegistry();
		final IEditorDescriptor descriptor = registry.getDefaultEditor( p_input.getName() );
		
		if ( descriptor != null ) {
			return descriptor.getId();
		}

		return null;
	}
	
	protected abstract String getEditorId( SDDebugBreakpoint p_sdBreakpoint );
	
	/**
	 * @see IDebugModelPresentation#getEditorInput(Object)
	 */
	@Override
	public IEditorInput getEditorInput( final Object p_element ) {
		if ( p_element instanceof IFile ) {
			return new FileEditorInput( (IFile) p_element );
		}
		
		if( p_element instanceof LocalFileStorage ) {
			final LocalFileStorage lfc = (LocalFileStorage) p_element;
			
			return new StorageEditorInput( lfc, lfc.getFile() );
		}
		
		if ( p_element instanceof SDDebugBreakpoint ) {
			try {
				return getEditorInput( (SDDebugBreakpoint) p_element );
			} 
			catch ( final CoreException p_ex ) {
				throw new RuntimeException( p_ex );
			}
		}
		
		return null;
	}
	
	protected abstract IEditorInput getEditorInput( SDDebugBreakpoint p_sdBreakpoint )
	throws CoreException;
	
	/**
	 * @see IDebugModelPresentation#getImage(Object)
	 */
	@Override
	public Image getImage( final Object p_element ) {
		
		// Return null to ensure standard behavior
		return null;
	}
	
	/**
	 * @see IDebugModelPresentation#getText(Object)
	 */
	@Override
	public String getText( final Object p_element ) {
		if ( p_element instanceof SDDebugBreakpoint ) {
			return getBreakpointText( (SDDebugBreakpoint) p_element );
		} 

		// Return null to ensure standard behavior
		return null;
	}

	/**
	 * @see IDebugModelPresentation#computeDetail(IValue, IValueDetailListener)
	 */
	@Override
	public void computeDetail(	final IValue p_value,
								final IValueDetailListener p_listener ) {
		String detail = "";
		
		try {
			detail = p_value.getValueString();
		} 
		catch ( final DebugException p_ex ) {
		}

		p_listener.detailComputed( p_value, detail );
	}

	private String getBreakpointText( final SDDebugBreakpoint p_breakpoint ) {
		final StringBuilder label = new StringBuilder();
		label.append( p_breakpoint.getElementName() );
		label.append(" [");
		label.append( p_breakpoint.getElementURI() );
		label.append( "] ");
		label.append(" [");
		label.append( p_breakpoint.getElementUUID() );
		label.append( "] ");

		return label.toString();
	}

    class StorageEditorInput extends PlatformObject implements IStorageEditorInput {
    	
    	private File fFile;

    	private IStorage fStorage;

    	public StorageEditorInput( 	final IStorage p_storage, 
    								final File p_file ) {
    		super();
	
    		fStorage = p_storage;
	
    		fFile = p_file;
    	}

		@Override
		public IStorage getStorage() {
			return fStorage;
		}

		@Override
		public ImageDescriptor getImageDescriptor() {
			return null;
		}

		@Override
		public String getName() {
			return getStorage().getName();
		}

		@Override
		public IPersistableElement getPersistable() {
			return null;
		}

		@Override
		public String getToolTipText() {
			return getStorage().getFullPath().toOSString();
		}

		@Override
		public boolean equals( final Object p_object ) {
			return p_object instanceof StorageEditorInput && getStorage().equals( ( (StorageEditorInput) p_object ).getStorage() );
		}

		@Override
		public int hashCode() {
			return getStorage().hashCode();
		}

		@Override
		public boolean exists() {
			return fFile.exists();
		}
    }
}
